@@ -119,15 +119,56 @@ class MyCache(DistUpgrade.DistUpgradeCache.MyCache):
     def keep_count(self):
         return self._depcache.keep_count
 
+    def _check_dependencies(self, target, deps):
+        """Return True if any of the dependencies in deps match target."""
+        # TODO: handle virtual packages
+        for dep_or in deps:
+            if not dep_or:
+                continue
+            match = True
+            for base_dep in dep_or:
+                if (base_dep.name != target.package.shortname or
+                    not apt_pkg.check_dep(
+                        target.version, base_dep.relation, base_dep.version)):
+                    match = False
+            if match:
+                return True
+        return False
+
+    def find_removal_justification(self, pkg):
+        target = pkg.installed
+        if not target:
+            return False
+        for cpkg in self:
+            candidate = cpkg.candidate
+            if candidate is not None:
+                if (self._check_dependencies(
+                        target, candidate.get_dependencies("Conflicts")) and
+                    self._check_dependencies(
+                        target, candidate.get_dependencies("Replaces"))):
+                    logging.info(
+                        "%s Conflicts/Replaces %s; allowing removal" % (
+                            candidate.package.shortname, pkg.shortname))
+                    return True
+        return False
+
     def saveDistUpgrade(self):
         """ this functions mimics a upgrade but will never remove anything """
         #self._apply_dselect_upgrade()
         self._depcache.upgrade(True)
         wouldDelete = self._depcache.del_count
-        if self._depcache.del_count > 0:
+        if wouldDelete > 0:
+            deleted_pkgs = [pkg for pkg in self if pkg.marked_delete]
+            assert wouldDelete == len(deleted_pkgs)
+            for pkg in deleted_pkgs:
+                if self.find_removal_justification(pkg):
+                    wouldDelete -= 1
+        if wouldDelete > 0:
             self.clear()
-        assert (self._depcache.broken_count == 0 and
-                self._depcache.del_count == 0)
+            assert (self._depcache.broken_count == 0 and
+                    self._depcache.del_count == 0)
+        else:
+            assert self._depcache.broken_count == 0
         #self._apply_dselect_upgrade()
         self._depcache.upgrade()
         return wouldDelete
@@ -304,11 +345,9 @@ class MyCache(DistUpgrade.DistUpgradeCache.MyCache):
         origins = self[name].candidate.origins
         self.all_changes[name] = _("Changes for %s versions:\n"
                                    "Installed version: %s\n"
-                                   "Available version: %s\n\n") % (
-                                       name,
-                                       getattr(self[name].installed, "version",
-                                               None),
-                                       self[name].candidate.version)
+                                   "Available version: %s\n\n") % \
+            (name, getattr(self[name].installed, "version", None),
+             self[name].candidate.version)
         if not self.CHANGELOG_ORIGIN in [o.origin for o in origins]:
             self._fetch_changelog_for_third_party_package(name)
             return
