@@ -277,7 +277,9 @@ class NeedRestartDialog(InternalDialog):
         InternalDialog.__init__(self, window_main)
         self.set_header(
             _("The computer needs to restart to finish installing updates."))
-        self.focus_button = self.add_button(_("_Restart"), self.restart)
+        self.add_settings_button()
+        self.add_button(_("Restart _Later"), self.window_main.close)
+        self.focus_button = self.add_button(_("_Restart Now…"), self.restart)
 
     def start(self):
         Dialog.start(self)
@@ -288,6 +290,7 @@ class NeedRestartDialog(InternalDialog):
 
     def restart(self, *args, **kwargs):
         self._request_reboot_via_session_manager()
+        self.window_main.close()
 
     def _request_reboot_via_session_manager(self):
         try:
@@ -309,5 +312,18 @@ class NeedRestartDialog(InternalDialog):
             iface = dbus.Interface(
                 proxy_obj, "org.freedesktop.ConsoleKit.Manager")
             iface.Restart()
+        except dbus.DBusException:
+            self._request_reboot_via_logind()
+        except:
+            pass
+
+    def _request_reboot_via_logind(self):
+        try:
+            bus = dbus.SystemBus()
+            proxy_obj = bus.get_object("org.freedesktop.login1",
+                                       "/org/freedesktop/login1")
+            iface = dbus.Interface(
+                proxy_obj, "org.freedesktop.login1.Manager")
+            iface.Reboot(True)
         except dbus.DBusException:
             pass
