@@ -23,17 +23,21 @@
 from __future__ import absolute_import, print_function
 
 from gi.repository import Gtk
-from gi.repository import Gdk
+from gi.repository import Gdk, GdkX11
 from gi.repository import Gio
 from gi.repository import GLib
 
+GdkX11  # pyflakes
+
 import warnings
 warnings.filterwarnings("ignore", "Accessed deprecated property",
                         DeprecationWarning)
 
 import apt_pkg
 import os
+import subprocess
 import sys
+import time
 from gettext import gettext as _
 
 import dbus
@@ -50,22 +54,22 @@ from .Dialogs import (DistUpgradeDialog,
                       PartialUpgradeDialog,
                       UnsupportedDialog,
                       UpdateErrorDialog)
-from .InstallProgress import InstallProgress
 from .MetaReleaseGObject import MetaRelease
-from .UpdateProgress import UpdateProgress
 from .UpdatesAvailable import UpdatesAvailable
 from .Core.AlertWatcher import AlertWatcher
 from .Core.MyCache import MyCache
 from .Core.roam import NetworkManagerHelper
 from .Core.UpdateList import UpdateList
+from .backend import (InstallBackend,
+                      get_backend)
 
 # file that signals if we need to reboot
 REBOOT_REQUIRED_FILE = "/var/run/reboot-required"
 
 
 class UpdateManager(Gtk.Window):
-    """ This class is the main window and work flow controller.  Panes will add
-        themselves to the main window and it will morph between them."""
+    """ This class is the main window and work flow controller. The main
+        window will show panes, and it will morph between them. """
 
     def __init__(self, datadir, options):
         Gtk.Window.__init__(self)
@@ -82,16 +86,14 @@ class UpdateManager(Gtk.Window):
         # Basic GTK+ parameters
         self.set_title(_("Software Updater"))
         self.set_icon_name("system-software-update")
-        self.set_resizable(False)
         self.set_position(Gtk.WindowPosition.CENTER)
 
         # Keep window at a constant size
         ctx = self.get_style_context()
         ctx.connect("changed", lambda ctx: self.resize_to_standard_width())
-        self.resize_to_standard_width()
 
         # Signals
-        self.connect("delete-event", self.close)
+        self.connect("delete-event", self._on_close)
 
         self._setup_dbus()
 
@@ -110,6 +112,17 @@ class UpdateManager(Gtk.Window):
             self.options and self.options.devel_release,
             self.options and self.options.use_proposed)
 
+    def begin_user_resizable(self, stored_width=0, stored_height=0):
+        self.set_resizable(True)
+        if stored_width > 0 and stored_height > 0:
+            # There is a race here.  If we immediately resize, it often doesn't
+            # take.  Using idle_add helps, but we *still* occasionally don't
+            # restore the size correctly.  Help needed to track this down!
+            GLib.idle_add(lambda: self.resize(stored_width, stored_height))
+
+    def end_user_resizable(self):
+        self.set_resizable(False)
+
     def resize_to_standard_width(self):
         if self.get_resizable():
             return  # only size to a specific em if we are a static size
@@ -130,56 +143,78 @@ class UpdateManager(Gtk.Window):
         self.disconnect(self.initial_focus_id)
         return False
 
-    def push(self, pane, controller):
-        child = self.get_child()
-        if child is not None:
-            if self.controller and hasattr(self.controller, "save_state"):
-                self.controller.save_state()
-            child.destroy()
+    def _start_pane(self, pane):
+        if self.controller is not None:
+            self.controller.stop()
+            if isinstance(self.controller, Gtk.Widget):
+                self.controller.destroy()
+
+        self.controller = pane
+        self._look_ready()
+        self.end_user_resizable()
 
         if pane is None:
-            self.controller = None
             return
 
-        pane.reparent(self)
-        self.controller = controller
+        if isinstance(pane, Gtk.Widget):
+            self.add(pane)
+            pane.start()
+            self.show_all()
+        else:
+            pane.start()
+            self.hide()
 
-        # Reset state
-        self.set_resizable(False)
-        self.resize_to_standard_width()
-        self.set_sensitive(True)
-        if self.get_window() is not None:
-            self.get_window().set_cursor(None)
-            self.get_window().set_functions(Gdk.WMFunction.ALL)
+    def _on_close(self, widget, data=None):
+        self.close()
 
-        if self.controller and hasattr(self.controller, "restore_state"):
-            self.controller.restore_state()
-
-        pane.show()
-        self.show()
-
-    def close(self, widget, data=None):
+    def close(self):
         if not self.get_sensitive():
             return True
 
-        if self.controller and hasattr(self.controller, "close"):
-            # let controller handle it as they will
-            if self.controller.close():
-                return
-
+        if self.controller:
+            controller_close = self.controller.close()
+            if controller_close:
+                return controller_close
         self.exit()
 
     def exit(self):
         """ exit the application, save the state """
-        self.push(None, None)
+        self._start_pane(None)
         sys.exit(0)
 
+    def show_settings(self):
+        try:
+            apt_pkg.pkgsystem_unlock()
+        except SystemError:
+            pass
+        cmd = ["/usr/bin/software-properties-gtk",
+               "--open-tab", "2",
+               "--toplevel", "%s" % self.get_window().get_xid()
+               ]
+        self._look_busy()
+        try:
+            p = subprocess.Popen(cmd)
+        except OSError:
+            pass
+        else:
+            while p.poll() is None:
+                while Gtk.events_pending():
+                    Gtk.main_iteration()
+                time.sleep(0.05)
+        finally:
+            self.start_available()
+
     def start_update(self):
         if self.options.no_update:
             self.start_available()
             return
 
-        self._start_pane(UpdateProgress(self))
+        update_backend = get_backend(self, InstallBackend.ACTION_UPDATE)
+        self._start_pane(update_backend)
+
+    def start_install(self):
+        install_backend = get_backend(self, InstallBackend.ACTION_INSTALL)
+        self._start_pane(install_backend)
 
     def start_available(self, cancelled_update=False, error_occurred=False):
         self._look_busy()
@@ -212,10 +247,7 @@ class UpdateManager(Gtk.Window):
                 header = _("You stopped the check for updates.")
                 desc = _("Updated software is available from "
                          "a previous check.")
-            return UpdatesAvailable(self, header, desc)
-
-    def start_install(self):
-        self._start_pane(InstallProgress(self))
+            return UpdatesAvailable(self, header, desc, need_reboot)
 
     def start_error(self, is_update_error, header, desc):
         if is_update_error:
@@ -223,15 +255,17 @@ class UpdateManager(Gtk.Window):
         else:
             self._start_pane(ErrorDialog(self, header, desc))
 
-    def _start_pane(self, pane):
-        self._look_busy()
-        pane.main()
-
     def _look_busy(self):
         self.set_sensitive(False)
         if self.get_window() is not None:
             self.get_window().set_cursor(Gdk.Cursor.new(Gdk.CursorType.WATCH))
 
+    def _look_ready(self):
+        self.set_sensitive(True)
+        if self.get_window() is not None:
+            self.get_window().set_cursor(None)
+            self.get_window().set_functions(Gdk.WMFunction.ALL)
+
     def _check_meta_release(self):
         if self.meta_release is None:
             return None
@@ -251,7 +285,7 @@ class UpdateManager(Gtk.Window):
             return UnsupportedDialog(self, self.meta_release)
 
         # Check for new fresh release
-        settings = Gio.Settings("com.ubuntu.update-manager")
+        settings = Gio.Settings.new("com.ubuntu.update-manager")
         if (self.meta_release.new_dist and
                 (self.options.check_dist_upgrades or
                  settings.get_boolean("check-dist-upgrades"))):
@@ -305,12 +339,14 @@ class UpdateManager(Gtk.Window):
             self.start_error(True, header, desc)
 
         # Let the Gtk event loop breath if it hasn't had a chance.
-        while Gtk.events_pending():
-            Gtk.main_iteration()
+        def iterate():
+            while Gtk.events_pending():
+                Gtk.main_iteration()
+        iterate()
 
         self.update_list = UpdateList(self)
         try:
-            self.update_list.update(self.cache)
+            self.update_list.update(self.cache, eventloop_callback=iterate)
         except SystemError as e:
             header = _("Could not calculate the upgrade")
             desc = _("An unresolvable problem occurred while "
