@@ -36,7 +36,6 @@ from gi.repository import Gtk
 from gi.repository import Gdk
 from gi.repository import GObject
 from gi.repository import Gio
-GObject.threads_init()
 from gi.repository import Pango
 
 import warnings
@@ -48,22 +47,20 @@ import apt_pkg
 import os
 import re
 import logging
-import subprocess
 import time
 import threading
 
 from gettext import gettext as _
 from gettext import ngettext
 
-
-from .Core.utils import (get_package_label, humanize_size)
+from .Core.utils import humanize_size
 from .Core.AlertWatcher import AlertWatcher
 from .Core.UpdateList import UpdateSystemGroup
+from .Dialogs import InternalDialog
 
 from DistUpgrade.DistUpgradeCache import NotEnoughFreeSpaceError
 
 from .ChangelogViewer import ChangelogViewer
-from .SimpleGtkbuilderApp import SimpleGtkbuilderApp
 from .UnitySupport import UnitySupport
 
 
@@ -104,8 +101,10 @@ class CellAreaPackage(Gtk.CellAreaBox):
     def do_foreach_alloc(self, context, widget, cell_area_in, bg_area_in,
                          callback):
         cells = []
+
         def gather(cell, data):
             cells.append(cell)
+
         self.foreach(gather, None)
 
         cell_is_hidden = {}
@@ -203,34 +202,55 @@ class CellAreaPackage(Gtk.CellAreaBox):
             return cell_start
 
 
-class UpdatesAvailable(SimpleGtkbuilderApp):
+class UpdatesAvailable(InternalDialog):
+    APP_INSTALL_ICONS_PATH = "/usr/share/app-install/icons"
 
-    def __init__(self, app, header=None, desc=None):
-        self.window_main = app
-        self.datadir = app.datadir
-        self.options = app.options
-        self.cache = app.cache
-        self.list = app.update_list
-        SimpleGtkbuilderApp.__init__(
-            self,
-            self.datadir + "/gtkbuilder/UpdateManager.ui",
-            "update-manager")
+    def __init__(self, window_main, header=None, desc=None,
+                 need_reboot=False):
+        InternalDialog.__init__(self, window_main)
+
+        self.window_main = window_main
+        self.datadir = window_main.datadir
+        self.cache = window_main.cache
+
+        self.custom_header = header
+        self.custom_desc = desc
+        self.need_reboot = need_reboot
+
+        content_ui_path = os.path.join(self.datadir,
+                                       "gtkbuilder/UpdateManager.ui")
+        self._load_ui(content_ui_path, "pane_updates_available")
+        self.set_content_widget(self.pane_updates_available)
+
+        self.dl_size = 0
+        self.connected = True
 
         # Used for inhibiting power management
         self.sleep_cookie = None
         self.sleep_dev = None
 
-        # workaround for LP: #945536
-        self.clearing_store = False
+        self.settings = Gio.Settings.new("com.ubuntu.update-manager")
 
-        self.custom_header = header
-        self.custom_desc = desc
+        # Special icon theme for looking up app-install-data icons
+        self.app_icons = Gtk.IconTheme.get_default()
+        self.app_icons.append_search_path(self.APP_INSTALL_ICONS_PATH)
 
-        self.button_close.grab_focus()
-        self.dl_size = 0
-        self.connected = True
+        # Create Unity launcher quicklist
+        # FIXME: instead of passing parent we really should just send signals
+        self.unity = UnitySupport(parent=self)
 
-        self.settings = Gio.Settings("com.ubuntu.update-manager")
+        # setup the help viewer and disable the help button if there
+        # is no viewer available
+        #self.help_viewer = HelpViewer("update-manager")
+        #if self.help_viewer.check() == False:
+        #    self.button_help.set_sensitive(False)
+
+        self.add_settings_button()
+        self.button_close = self.add_button(Gtk.STOCK_CANCEL,
+                                            self.window_main.close)
+        self.button_install = self.add_button(_("Install Now"),
+                                              self.on_button_install_clicked)
+        self.focus_button = self.button_install
 
         # create text view
         self.textview_changes = ChangelogViewer()
@@ -239,27 +259,12 @@ class UpdatesAvailable(SimpleGtkbuilderApp):
         changes_buffer = self.textview_changes.get_buffer()
         changes_buffer.create_tag("versiontag", weight=Pango.Weight.BOLD)
 
-        # expander
-        self.expander_details.set_expanded(
-            self.settings.get_boolean("show-details"))
-        self.expander_details.connect("activate", self.pre_activate_details)
-        self.expander_details.connect("notify::expanded",
-                                      self.activate_details)
-        self.expander_desc.connect("notify::expanded", self.activate_desc)
-
-        # useful exit stuff
-        self.button_close.connect("clicked", lambda w: self.window_main.exit())
-
-        # If auto-updates are on, change cancel label
-        self.notifier_settings = Gio.Settings("com.ubuntu.update-notifier")
-        self.notifier_settings.connect(
-            "changed::auto-launch",
-            lambda s, p: self.update_close_button())
-        self.update_close_button()
-
         # the treeview (move into it's own code!)
         self.store = Gtk.TreeStore(str, GObject.TYPE_PYOBJECT, str, bool)
-        self.treeview_update.set_model(self.store)
+        self.treeview_update.set_model(None)
+
+        self.image_restart.set_from_gicon(self.get_restart_icon(),
+                                          Gtk.IconSize.BUTTON)
 
         restart_icon_renderer = Gtk.CellRendererPixbuf()
         restart_icon_renderer.set_property("xpad", 4)
@@ -273,9 +278,9 @@ class UpdatesAvailable(SimpleGtkbuilderApp):
         restart_column.set_cell_data_func(restart_icon_renderer,
                                           self.restart_icon_renderer_data_func)
 
-        pkg_area = CellAreaPackage(bool(self.list.security_groups))
-        pkg_column = Gtk.TreeViewColumn.new_with_area(pkg_area)
-        pkg_area.column = pkg_column
+        self.pkg_cell_area = CellAreaPackage(False)
+        pkg_column = Gtk.TreeViewColumn.new_with_area(self.pkg_cell_area)
+        self.pkg_cell_area.column = pkg_column
         pkg_column.set_title(_("Install"))
         pkg_column.set_property("spacing", 4)
         pkg_column.set_expand(True)
@@ -324,24 +329,26 @@ class UpdatesAvailable(SimpleGtkbuilderApp):
         self.treeview_update.connect("button-press-event",
                                      self.on_treeview_button_press)
 
-        # setup the help viewer and disable the help button if there
-        # is no viewer available
-        #self.help_viewer = HelpViewer("update-manager")
-        #if self.help_viewer.check() == False:
-        #    self.button_help.set_sensitive(False)
-
-        if not os.path.exists("/usr/bin/software-properties-gtk"):
-            self.button_settings.set_sensitive(False)
-
         # init show version
         self.show_versions = self.settings.get_boolean("show-versions")
         # init summary_before_name
         self.summary_before_name = self.settings.get_boolean(
             "summary-before-name")
 
-        # Create Unity launcher quicklist
-        # FIXME: instead of passing parent we really should just send signals
-        self.unity = UnitySupport(parent=self)
+        # expander
+        self.expander_details.set_expanded(
+            self.settings.get_boolean("show-details"))
+        self.expander_details.connect("activate", self.pre_activate_details)
+        self.expander_details.connect("notify::expanded",
+                                      self.activate_details)
+        self.expander_desc.connect("notify::expanded", self.activate_desc)
+
+        # If auto-updates are on, change cancel label
+        self.notifier_settings = Gio.Settings.new("com.ubuntu.update-notifier")
+        self.notifier_settings.connect(
+            "changed::auto-launch",
+            lambda s, p: self.update_close_button())
+        self.update_close_button()
 
         # Alert watcher
         self.alert_watcher = AlertWatcher()
@@ -350,11 +357,20 @@ class UpdatesAvailable(SimpleGtkbuilderApp):
         self.alert_watcher.connect("network-3g-alert",
                                    self._on_network_3g_alert)
 
+    def stop(self):
+        InternalDialog.stop(self)
+        self._save_state()
+
+    def start(self):
+        InternalDialog.start(self)
+        self.set_update_list(self.window_main.update_list)
+        self.alert_watcher.check_alert_state()
+        self._restore_state()
+
     def is_auto_update(self):
         update_days = apt_pkg.config.find_i(
             "APT::Periodic::Update-Package-Lists")
-        auto_launch = self.notifier_settings.get_boolean("auto-launch")
-        return update_days >= 1 and auto_launch
+        return update_days >= 1
 
     def update_close_button(self):
         if self.is_auto_update():
@@ -368,41 +384,47 @@ class UpdatesAvailable(SimpleGtkbuilderApp):
 
     def install_all_updates(self, menu, menuitem, data):
         self.select_all_upgrades(None)
-        self.on_button_install_clicked(None)
+        self.on_button_install_clicked()
+
+    def pkg_requires_restart(self, pkg):
+        if pkg is None or pkg.candidate is None:
+            return False
+        restart_condition = pkg.candidate.record.get('Restart-Required')
+        return restart_condition == 'system'
+
+    def get_restart_icon(self):
+        # FIXME: Non-standard, incorrect icon name (from app category).
+        # Theme support for what we want seems to be lacking.
+        restart_icon_names = ['view-refresh-symbolic',
+                              'system-restart',
+                              'system-reboot']
+        return Gio.ThemedIcon.new_from_names(restart_icon_names)
 
     def restart_icon_renderer_data_func(self, cell_layout, renderer, model,
-                                        iter, data):
-        def pkg_requires_restart(pkg):
-            restart_condition = pkg.candidate.record.get('XB-Restart-Required')
-            return restart_condition == 'system'
-
+                                        iter, user_data):
         data = model.get_value(iter, LIST_UPDATE_DATA)
         path = model.get_path(iter)
 
         requires_restart = False
-        if data.item:
-            requires_restart = pkg_requires_restart(data.item.pkg)
+        if data.item and data.item.pkg:
+            requires_restart = self.pkg_requires_restart(data.item.pkg)
         elif data.group:
             if not self.treeview_update.row_expanded(path):
                 # A package in the group requires restart
                 for group_item in data.group.items:
-                    if pkg_requires_restart(group_item.pkg):
+                    if group_item.pkg and self.pkg_requires_restart(
+                            group_item.pkg):
                         requires_restart = True
                         break
 
-        # FIXME: Non-standard, incorrect icon name (from app category).
-        # Theme support for what we want seems to be lacking.
         if requires_restart:
-            restart_icon_names = ['view-refresh-symbolic',
-                                  'system-restart',
-                                  'system-reboot']
-            gicon = Gio.ThemedIcon.new_from_names(restart_icon_names)
+            gicon = self.get_restart_icon()
         else:
             gicon = None
         renderer.set_property("gicon", gicon)
 
-    def pkg_toggle_renderer_data_func(self, cell_layout, renderer, model, iter,
-                                      data):
+    def pkg_toggle_renderer_data_func(self, cell_layout, renderer, model,
+                                      iter, user_data):
         data = model.get_value(iter, LIST_UPDATE_DATA)
 
         activatable = False
@@ -419,8 +441,7 @@ class UpdatesAvailable(SimpleGtkbuilderApp):
             saw_install = None
             for group in data.groups:
                 for item in group.items:
-                    pkg = item.pkg
-                    this_install = pkg.marked_install or pkg.marked_upgrade
+                    this_install = item.is_selected()
                     if saw_install is not None and saw_install != this_install:
                         inconsistent = True
                         break
@@ -433,20 +454,35 @@ class UpdatesAvailable(SimpleGtkbuilderApp):
         renderer.set_property("activatable", activatable)
         renderer.set_property("inconsistent", inconsistent)
 
-    def pkg_icon_renderer_data_func(self, cell_layout, renderer, model, iter,
-                                    data):
+    def pkg_icon_renderer_data_func(self, cell_layout, renderer, model,
+                                    iter, user_data):
         data = model.get_value(iter, LIST_UPDATE_DATA)
 
         gicon = None
         if data.group:
-            gicon = data.group.icon
+            gicon = self.get_app_install_icon(data.group.icon)
         elif data.item:
-            gicon = data.item.icon
+            gicon = self.get_app_install_icon(data.item.icon)
 
         renderer.set_property("gicon", gicon)
 
-    def pkg_label_renderer_data_func(self, cell_layout, renderer, model, iter,
-                                     data):
+    def get_app_install_icon(self, icon):
+        """Any application icon is coming from app-install-data's desktop
+           files, which refer to icons from app-install-data's icon directory.
+           So we look them up here."""
+
+        if not isinstance(icon, Gio.ThemedIcon):
+            return icon  # shouldn't happen
+
+        info = self.app_icons.choose_icon(icon.get_names(), 16,
+                                          Gtk.IconLookupFlags.FORCE_SIZE)
+        if info is not None:
+            return Gio.FileIcon.new(Gio.File.new_for_path(info.get_filename()))
+        else:
+            return icon  # Assume it's in one of the user's themes
+
+    def pkg_label_renderer_data_func(self, cell_layout, renderer, model,
+                                     iter, user_data):
         data = model.get_value(iter, LIST_UPDATE_DATA)
         name = GLib.markup_escape_text(model.get_value(iter, LIST_NAME))
 
@@ -497,7 +533,7 @@ class UpdatesAvailable(SimpleGtkbuilderApp):
         if (item is None and data.group is not None and
                 data.group.core_item is not None):
             item = data.group.core_item
-        if (item is None or item.pkg is None or 
+        if (item is None or item.pkg is None or
                 item.pkg.candidate is None or
                 item.pkg.candidate.description is None):
             changes_buffer = self.textview_changes.get_buffer()
@@ -658,8 +694,7 @@ class UpdatesAvailable(SimpleGtkbuilderApp):
             data = model.get_value(iter, LIST_UPDATE_DATA)
             active = False
             if data.item:
-                active = data.item.pkg.marked_install or \
-                         data.item.pkg.marked_upgrade
+                active = data.item.is_selected()
             elif data.group:
                 active = data.group.packages_are_selected()
             elif data.groups:
@@ -667,6 +702,26 @@ class UpdatesAvailable(SimpleGtkbuilderApp):
             model.set_value(iter, LIST_TOGGLE_ACTIVE, active)
         self.store.foreach(foreach_cb, None)
 
+    def _check_for_required_restart(self):
+        requires_restart = False
+
+        def foreach_cb(model, path, iter, data):
+            data = model.get_value(iter, LIST_UPDATE_DATA)
+            active = model.get_value(iter, LIST_TOGGLE_ACTIVE)
+            if not active:
+                return
+            pkg = None
+            if data.item:
+                pkg = data.item.pkg
+            elif data.group and data.group.core_item:
+                pkg = data.group.core_item.pkg
+            if pkg and self.pkg_requires_restart(pkg):
+                nonlocal requires_restart
+                requires_restart = True
+
+        self.store.foreach(foreach_cb, None)
+        self.hbox_restart.set_visible(requires_restart)
+
     def _refresh_updates_count(self):
         self.button_install.set_sensitive(self.cache.install_count)
         try:
@@ -710,10 +765,11 @@ class UpdatesAvailable(SimpleGtkbuilderApp):
 
     def updates_changed(self):
         self._mark_selected_updates()
+        self._check_for_required_restart()
         self._refresh_updates_count()
 
     def update_count(self):
-        """activate or disable widgets and show dialog texts correspoding to
+        """activate or disable widgets and show dialog texts corresponding to
            the number of available updates"""
         self.updates_changed()
 
@@ -722,6 +778,9 @@ class UpdatesAvailable(SimpleGtkbuilderApp):
 
         if self.custom_header is not None:
             text_header = self.custom_header
+        if self.custom_desc is not None:
+            text_desc = self.custom_desc
+
         # show different text on first run (UX team suggestion)
         elif self.settings.get_boolean("first-run"):
             flavor = self.window_main.meta_release.flavor_name
@@ -733,20 +792,14 @@ class UpdatesAvailable(SimpleGtkbuilderApp):
         else:
             text_header = _("Updated software is available for this "
                             "computer. Do you want to install it now?")
-
-        if self.custom_desc is not None:
-            text_desc = self.custom_desc
+            if not self.hbox_restart.get_visible() and self.need_reboot:
+                text_desc = _("The computer also needs to restart "
+                              "to finish installing previous updates.")
 
         self.notebook_details.set_sensitive(True)
         self.treeview_update.set_sensitive(True)
-        self.button_install.grab_default()
-        self.label_header.set_label(text_header)
-
-        if text_desc is not None:
-            self.label_desc.set_label(text_desc)
-            self.label_desc.show()
-        else:
-            self.label_desc.hide()
+        self.set_header(text_header)
+        self.set_desc(text_desc)
 
         return True
 
@@ -754,47 +807,20 @@ class UpdatesAvailable(SimpleGtkbuilderApp):
     def pre_activate_details(self, expander):
         expanded = self.expander_details.get_expanded()
         if expanded:
-            self.save_state()
+            self._save_state()
 
     def activate_details(self, expander, data):
         expanded = self.expander_details.get_expanded()
         self.settings.set_boolean("show-details", expanded)
         if expanded:
             self.on_treeview_update_cursor_changed(self.treeview_update)
-        self.restore_state()
+        self._restore_state()
 
     def activate_desc(self, expander, data):
         expanded = self.expander_desc.get_expanded()
         self.expander_desc.set_vexpand(expanded)
 
-    #def on_button_help_clicked(self, widget):
-    #  self.help_viewer.run()
-
-    def on_button_settings_clicked(self, widget):
-        #print("on_button_settings_clicked")
-        try:
-            apt_pkg.pkgsystem_unlock()
-        except SystemError:
-            pass
-        cmd = ["/usr/bin/software-properties-gtk",
-               "--open-tab", "2",
-               # FIXME: once get_xid() is available via introspections, add
-               #        this back
-               #"--toplevel", "%s" % self.window_main.get_window().get_xid()
-               ]
-        self.window_main.set_sensitive(False)
-        p = subprocess.Popen(cmd)
-        while p.poll() is None:
-            while Gtk.events_pending():
-                Gtk.main_iteration()
-            time.sleep(0.05)
-        apt_pkg.init_config()
-        self.window_main.refresh_cache()
-        self.fillstore()
-        self.update_close_button()
-        self.window_main.set_sensitive(True)
-
-    def on_button_install_clicked(self, widget):
+    def on_button_install_clicked(self):
         self.unity.set_install_menuitem_visible(False)
         #print("on_button_install_clicked")
         err_sum = _("Not enough free disk space")
@@ -871,32 +897,13 @@ class UpdatesAvailable(SimpleGtkbuilderApp):
         data = self.store.get_value(iter, LIST_UPDATE_DATA)
         # make sure that we don't allow to toggle deactivated updates
         # this is needed for the call by the row activation callback
-        if data.groups or data.group:
-            if data.groups:
-                self.toggle_from_groups(data.groups)
-            elif data.group:
-                self.toggle_from_groups([data.group])
-            self.updates_changed()
-            self.treeview_update.queue_draw()
-            return
-        if (data.item is None or data.item.pkg is None or
-                data.item.pkg.name in self.list.held_back):
-            return False
-        self.setBusy(True)
-        # update the cache
-        if data.item.pkg.marked_install or data.item.pkg.marked_upgrade:
-            data.item.pkg.mark_keep()
-            if self.cache._depcache.broken_count:
-                Fix = apt_pkg.ProblemResolver(self.cache._depcache)
-                Fix.resolve_by_keep()
+        if data.groups:
+            self.toggle_from_items([item for group in data.groups
+                                    for item in group.items])
+        elif data.group:
+            self.toggle_from_items(data.group.items)
         else:
-            try:
-                data.item.pkg.mark_install()
-            except SystemError:
-                pass
-        self.updates_changed()
-        self.treeview_update.queue_draw()
-        self.setBusy(False)
+            self.toggle_from_items([data.item])
 
     def on_treeview_update_row_activated(self, treeview, path, column, *args):
         """
@@ -905,18 +912,20 @@ class UpdatesAvailable(SimpleGtkbuilderApp):
         """
         self.on_update_toggled(None, path)
 
-    def toggle_from_groups(self, groups):
+    def toggle_from_items(self, items):
         self.setBusy(True)
         actiongroup = apt_pkg.ActionGroup(self.cache._depcache)
 
-        # Deselect all updates if any in group are selected
-        keep_packages = any([g.packages_are_selected() for g in groups])
-        for group in groups:
-            for item in group.items:
+        # Deselect all updates if any are selected
+        keep_packages = any([item.is_selected() for item in items])
+        for item in items:
+            try:
                 if keep_packages:
                     item.pkg.mark_keep()
                 elif not item.pkg.name in self.list.held_back:
-                    item.pkg.mark_install(auto_fix=False, auto_inst=False)
+                    item.pkg.mark_install()
+            except SystemError:
+                pass
 
         # check if we left breakage
         if self.cache._depcache.broken_count:
@@ -927,24 +936,22 @@ class UpdatesAvailable(SimpleGtkbuilderApp):
         del actiongroup
         self.setBusy(False)
 
-    def save_state(self):
+    def _save_state(self):
         """ save the state  (window-size for now) """
         if self.expander_details.get_expanded():
             (w, h) = self.window_main.get_size()
             self.settings.set_int("window-width", w)
             self.settings.set_int("window-height", h)
 
-    def restore_state(self):
+    def _restore_state(self):
         """ restore the state (window-size for now) """
         w = self.settings.get_int("window-width")
         h = self.settings.get_int("window-height")
         expanded = self.expander_details.get_expanded()
-        self.window_main.set_resizable(expanded)
-        if w > 0 and h > 0 and expanded:
-            # There is a race here.  If we immediately resize, it often doesn't
-            # take.  Using idle_add helps, but we *still* occasionally don't
-            # restore the size correctly.  Help needed to track this down!
-            GLib.idle_add(lambda: self.window_main.resize(w, h))
+        if expanded:
+            self.window_main.begin_user_resizable(w, h)
+        else:
+            self.window_main.end_user_resizable()
         return False
 
     def _add_header(self, name, groups):
@@ -993,7 +1000,9 @@ class UpdatesAvailable(SimpleGtkbuilderApp):
                 ]
                 self.store.append(group_iter, item_row)
 
-    def fillstore(self):
+    def set_update_list(self, update_list):
+        self.list = update_list
+
         # use the watch cursor
         self.setBusy(True)
         # disconnect the view first
@@ -1014,14 +1023,11 @@ class UpdatesAvailable(SimpleGtkbuilderApp):
             self._add_groups(self.list.update_groups)
 
         self.treeview_update.set_model(self.store)
+        self.pkg_cell_area.indent_toplevel = bool(self.list.security_groups)
+        self.update_close_button()
         self.update_count()
         self.setBusy(False)
         while Gtk.events_pending():
             Gtk.main_iteration()
         self.updates_changed()
         return False
-
-    def main(self):
-        self.window_main.push(self.pane_updates_available, self)
-        self.fillstore()
-        self.alert_watcher.check_alert_state()
